import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;
import javax.swing.text.*;

public class SpaceviewTextArea extends JTextPane{

	/**
	 * 󔒉eLXgGAihttp://terai.xrea.jp/Swing/WhitespaceMark.htmlj
	 */
	
	private boolean wrap = false;
	private MyEditorKit editorKit = new MyEditorKit();
	
	public SpaceviewTextArea() {
		super();
		
		setEditorKit(editorKit);
	}
	
	public void setSpaceView(boolean setSpace) {
		editorKit.setSpaceView(setSpace);
	}
	
	public void setLineWrap(boolean setWrap) {
		wrap = setWrap;
	}
	
	public boolean getLineWrap() {
		return wrap;
	}
	
	public int getLineOfOffset(int offset) throws BadLocationException {
		Document doc = getDocument();
		if (offset < 0) {	throw new BadLocationException("Can't translate offset to line", -1);	}
		else if (offset > doc.getLength()) {	throw new BadLocationException("Can't translate offset to line", doc.getLength()+1);	}
		else {
			Element map = getDocument().getDefaultRootElement();
            		return map.getElementIndex(offset);
		}
	}
	
	public int getLineCount() {
		Element map = getDocument().getDefaultRootElement();
		return map.getElementCount();
	}
	
	public int getLineStartOffset(int line) throws BadLocationException {
		int lineCount = getLineCount();
		if (line < 0) {	throw new BadLocationException("Negative line", -1);	}
		else if (line >= lineCount) {	throw new BadLocationException("No such line", getDocument().getLength()+1);	}
		else {
			Element map = getDocument().getDefaultRootElement();
			Element lineElem = map.getElement(line);
			return lineElem.getStartOffset();
		}
	}
	
	public boolean getScrollableTracksViewportWidth() {
		if (wrap) {	return super.getScrollableTracksViewportWidth();	}
		Object parent = getParent();
		if (parent instanceof JViewport) {
			JViewport port = (JViewport) parent;
			int w = port.getWidth();	// \ł͈()
			TextUI ui = getUI();
			Dimension sz = ui.getPreferredSize(this); // ۂ̕TCY
			if (sz.width < w) {
				return true;
			}
		}
		return false;
	}
	
	private void setEditorKit(MyEditorKit kit) {
		((JTextPane) ((JTextComponent) this)).setEditorKit(kit);
	}

}

class MyEditorKit extends StyledEditorKit {

	private final SimpleAttributeSet attrs = new SimpleAttributeSet();
	boolean spaceview = false;
	public void install(JEditorPane c) {
		FontMetrics fm = c.getFontMetrics(c.getFont());
		int tabLength = fm.charWidth('m') * 4;
		TabStop[] tabs = new TabStop[100];
		for(int j=0;j<tabs.length;j++) {
			tabs[j] = new TabStop((j+1)*tabLength);
		}
		TabSet tabSet = new TabSet(tabs);
		StyleConstants.setTabSet(attrs, tabSet);
		super.install(c);
	}
	public ViewFactory getViewFactory() {
		return new MyViewFactory(spaceview);
	}
	public Document createDefaultDocument() {
		Document d = super.createDefaultDocument();
		if(d instanceof StyledDocument) {
			((StyledDocument)d).setParagraphAttributes(0, d.getLength(), attrs, false);
		}
		return d;
	}
	public void setSpaceView(boolean setSpace) {
		spaceview = setSpace;
	}

}
class MyViewFactory implements ViewFactory {

	private boolean setSpaceView = false;
	public MyViewFactory(boolean setSpace) {
		super();
		setSpaceView = setSpace;
	}
	public View create(Element elem) {
		String kind = elem.getName();
		if(kind!=null) {
			if(kind.equals(AbstractDocument.ContentElementName)) {
				return new WhitespaceLabelView(elem, setSpaceView);
			}else if(kind.equals(AbstractDocument.ParagraphElementName)) {
				return new MyParagraphView(elem, setSpaceView);
			}else if(kind.equals(AbstractDocument.SectionElementName)) {
				return new BoxView(elem, View.Y_AXIS);
			}else if(kind.equals(StyleConstants.ComponentElementName)) {
				return new ComponentView(elem);
			}else if(kind.equals(StyleConstants.IconElementName)) {
				return new IconView(elem);
			}
		}
		return new WhitespaceLabelView(elem);
	}
}

class MyParagraphView extends ParagraphView {

	private static final Color pc = new Color(120, 130, 110);
	private boolean setSpaceView = false;
	public MyParagraphView(Element elem) {
		super(elem);
	}
	public MyParagraphView(Element elem, boolean setSpace) {
		super(elem);
		setSpaceView = setSpace;
	}
	@Override public void paint(Graphics g, Shape allocation) {
		super.paint(g, allocation);
		if (setSpaceView) {	paintCustomParagraph(g, allocation);	}
	}
	private void paintCustomParagraph(Graphics g, Shape a) {
		try{
			Shape paragraph = modelToView(getEndOffset(), a, Position.Bias.Backward);
			Rectangle r = (paragraph==null)?a.getBounds():paragraph.getBounds();
			int x = r.x;
			int y = r.y;
			int h = r.height;
			Color old = g.getColor();
			g.setColor(pc);
			g.drawLine(x+1, y+h/2, x+1, y+h-4);
			g.drawLine(x+2, y+h/2, x+2, y+h-5);
			g.drawLine(x+3, y+h-6, x+3, y+h-6);
			g.setColor(old);
		}catch(Exception e) { e.printStackTrace(); }
	}

}

class WhitespaceLabelView extends LabelView {

	private static final Color pc = new Color(130, 140, 120);
	private static final BasicStroke dashed = new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, new float[] {1.0f}, 0.0f);
	private boolean setSpaceView = false;
	public WhitespaceLabelView(Element elem) {
		super(elem);
	}
	public WhitespaceLabelView(Element elem, boolean setSpace) {
		super(elem);
		setSpaceView = setSpace;
	}
	@Override public void paint(Graphics g, Shape a) {
		super.paint(g,a);
		if (!setSpaceView) {	return;	}
		Graphics2D g2 = (Graphics2D)g;
		Stroke stroke = g2.getStroke();
		Rectangle alloc = (a instanceof Rectangle) ? (Rectangle)a : a.getBounds();
		FontMetrics fontMetrics = g.getFontMetrics();
		int spaceWidth = fontMetrics.stringWidth(" ");
		int doubleSpaceWidth = fontMetrics.stringWidth("@");
		int sumOfTabs  = 0;
		String text = getText(getStartOffset(),getEndOffset()).toString();
		for(int i=0;i<text.length();i++) {
			String s = text.substring(i,i+1);
			int previousStringWidth = fontMetrics.stringWidth(text.substring(0,i)) + sumOfTabs;
			int sx = alloc.x+previousStringWidth;
			int sy = alloc.y+alloc.height-fontMetrics.getDescent();
			if(" ".equals(s)) {
				g2.setStroke(dashed);
				g2.setPaint(pc);
				g2.drawLine(sx+1, sy-1, sx+spaceWidth-2, sy-1);
				g2.drawLine(sx+2,   sy, sx+spaceWidth-2, sy);
			} else if("@".equals(s)) {
				g2.setStroke(dashed);
				g2.setPaint(pc);
				g2.drawRect(sx+1, sy-fontMetrics.getHeight()+2, doubleSpaceWidth-2, fontMetrics.getHeight()-2);
			}else if("\t".equals(s)) {
				int tabWidth = (int)getTabExpander().nextTabStop((float)sx, i)-sx;
				g2.setColor(pc);
				g2.drawLine(sx+2, sy-0, sx+2+2, sy-0);
				g2.drawLine(sx+2, sy-1, sx+2+1, sy-1);
				g2.drawLine(sx+2, sy-2, sx+2+0, sy-2);
				g2.setStroke(dashed);
				g2.drawLine(sx+2, sy, sx+tabWidth-2, sy);
				sumOfTabs+=tabWidth;
			}
			g2.setStroke(stroke);
		}
	}

}
